"""Base client used by fetch and search clients."""
from abc import ABC, abstractmethod
from typing import Dict, List, NamedTuple, Union

import pkg_resources
from requests import Response as RequestsResponse
from requests.exceptions import HTTPError

from pybomb.exceptions import (
    BadRequestException,
    InvalidFilterFieldException,
    InvalidResponseException,
    InvalidReturnFieldException,
    InvalidSortFieldException,
)
from pybomb.response import Response


class ResponseParam(NamedTuple):
    """Control structure for marking fields as filter-able and sort-able."""

    is_filter: bool
    is_sort: bool


class Client(ABC):
    """Base class for GB API resource clients."""

    URI_BASE = "http://www.giantbomb.com/api/"

    RESPONSE_FIELD_MAP: Dict[str, ResponseParam] = {}
    RESPONSE_STATUS_OK = 1
    RESPONSE_FORMAT_JSON = "json"

    RESOURCE_NAME = ""

    SORT_ORDER_ASCENDING = "asc"
    SORT_ORDER_DESCENDING = "desc"

    def __init__(self, api_key: str) -> None:
        """Init Client with GB API key and default_response_format.

        Args:
            api_key: The GB API key to use for each request
        """
        self.api_key = api_key
        self._headers = {
            "User-Agent": f'Pybomb {pkg_resources.require("pybomb")[0].version}'
        }

    def _validate_return_fields(self, return_fields: List[str]) -> None:
        """Validate the given return fields against those allowed on the resource.

        Args:
            return_fields: Requested return fields

        Raises:
            InvalidReturnFieldException: Invalid return fields
                requested for the resource.
        """
        for return_field in return_fields:
            if return_field not in self.RESPONSE_FIELD_MAP:
                raise InvalidReturnFieldException(
                    f'"{return_field}" is an invalid return field'
                )

    def _validate_sort_field(self, sort_by: str) -> None:
        """Validate the given sort field against those allowed on the resource.

        Args:
            sort_by: The field to sort the response by

        Raises:
            InvalidSortFieldException: Invalid sort supplied for the resource
        """
        if (
            sort_by not in self.RESPONSE_FIELD_MAP
            or not self.RESPONSE_FIELD_MAP[sort_by].is_sort
        ):
            raise InvalidSortFieldException(f'"{sort_by}" is an invalid sort field')

    def _validate_filter_fields(self, filter_by: Dict[str, Union[str, int]]) -> None:
        """Validate the given filter fields against those allowed on the resource.

        Args:
            filter_by: Requested filter fields

        Raises:
            InvalidFilterFieldException: Invalid filter fields requested
                for the resource
        """
        for filter_field in filter_by:
            if (
                filter_field not in self.RESPONSE_FIELD_MAP
                or not self.RESPONSE_FIELD_MAP[filter_field].is_filter
            ):
                raise InvalidFilterFieldException(
                    f'"{filter_field}" is an invalid filter field'
                )

    @staticmethod
    def _create_search_filter(filter_by: Dict[str, Union[str, int]]) -> str:
        """Create a filter string to be used for the request using the supplied filters.

        Args:
            filter_by: Requested filter fields

        Returns:
            A string containing the filters in the format required by GB API
        """
        return ",".join(
            [f"{key}:{value}" for key, value in filter_by.items() if value is not None]
        )

    def _query(self, params: Dict[str, Union[str, int]]) -> Response:
        """Add required params, call GB API and format the response.

        Args:
            params: All of the params requested for the call

        Returns:
            A Response object containing the GB API response
        """
        params["api_key"] = self.api_key
        params["format"] = self.RESPONSE_FORMAT_JSON

        response = self._query_api(params)
        self._validate_response(response)

        return Response.from_response_data(response)

    @abstractmethod
    def _query_api(self, params: Dict[str, Union[str, int]]) -> RequestsResponse:
        """Handle actual query to GB API.

        Args:
            params: All requests and required resource query parameters

        Returns:
            The raw requests Response from the GB call
        """
        return RequestsResponse()  # pragma: no cover

    def _validate_response(self, response: RequestsResponse) -> None:
        """Validate the response from the GB API.

        Args:
            response: The raw requests response from the GB call

        Raises:
            InvalidResponseException: The response was invalid
            BadRequestException: The request to the GB API was invalid
        """
        try:
            response.raise_for_status()
        except HTTPError as http_error:
            raise BadRequestException(str(http_error))

        response_data = response.json()
        if response_data["status_code"] != self.RESPONSE_STATUS_OK:
            raise InvalidResponseException(
                f'Response code {response_data["status_code"]}: {response_data["error"]}'
            )



"""Base client to extend to create search clients for endpoints of the GiantBomb API."""
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Union

from requests import get, Response as RequestsResponse

from pybomb.clients.base.client import Client
from pybomb.response import Response


class SearchClient(Client):
    """Base class for search GB API resource clients."""

    def search(
        self,
        filter_by: Dict[str, Any],
        return_fields: List = None,
        sort_by: Optional[str] = None,
        desc: bool = True,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Response:
        """Full search of resource.

        Supports all search fields available in API
        http://www.giantbomb.com/api/documentation

        Args:
            filter_by: A map of fields to filter the search by. These will
                be validated against the availiable search fields
            return_fields: A list of fields to be returned by the response.
                These will be validated against the availiable return fields.
                The default is to return everything
            sort_by: The field to sort the items in the reponse by.
                These will be validated against the availiable sort fields.
            desc: If sort direction is DESC or not (ASC). Defaults to True
            limit: The max number of items to request
            offset: The start offset for the return items, based on the given sort.

        Returns:
             A PyBomb Response containing the results of the search
        """
        self._validate_filter_fields(filter_by)
        search_filter = self._create_search_filter(filter_by)

        search_params: Dict[str, Union[str, int]] = {"filter": search_filter}
        self._apply_return_fields(return_fields, search_params)
        self._apply_sort_by(sort_by, desc, search_params)
        self._apply_limit(limit, search_params)
        self._apply_offset(offset, search_params)

        response = self._query(search_params)

        return response

    @abstractmethod
    def quick_search(
        self,
        name: str,
        sort_by: Optional[str] = None,
        desc: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        """Search with a simplier API."""
        ...  # pragma: no cover

    def _apply_return_fields(
        self, return_fields: Optional[List], search_params: Dict[str, Union[str, int]],
    ) -> None:
        """Apply return filters to search params if any were supplied.

        Args:
            return_fields: A list of fields to be returned by the response.
                These will be validated against the availiable return fields.
                The default is to return everything
            search_params: A dictionary containing all search params
        """
        if return_fields is None:
            return

        self._validate_return_fields(return_fields)
        field_list = ",".join(return_fields)

        search_params["field_list"] = field_list

    def _apply_sort_by(
        self,
        sort_by: Optional[str],
        desc: bool,
        search_params: Dict[str, Union[str, int]],
    ) -> None:
        """Apply sort by to search params if one was supplied.

        Args:
            sort_by: The field to sort the items in the reponse by.
                These will be validated against the availiable sort fields.
            desc: If sort direction is DESC or not (ASC). Defaults to True
            search_params: A dictionary containing all search params
        """
        if sort_by is None:
            return

        self._validate_sort_field(sort_by)

        if desc:
            direction = self.SORT_ORDER_DESCENDING
        else:
            direction = self.SORT_ORDER_ASCENDING

        search_params["sort"] = f"{sort_by}:{direction}"

    def _apply_limit(
        self, limit: Optional[int], search_params: Dict[str, Union[str, int]],
    ) -> None:
        """Apply the limit to the search params if one was supplied.

        Args:
            limit: The max number of items to request
            search_params: A dictionary containing all search params

        """
        if limit is None:
            return

        search_params["limit"] = int(limit)

    def _apply_offset(
        self, offset: Optional[int], search_params: Dict[str, Union[str, int]],
    ) -> None:
        """Add the offset to search params if one was supplied.

        Args:
            offset: The start offset for the return items, based on the given sort.
            search_params: A dictionary containing all search params
        """
        if offset is None:
            return

        search_params["offset"] = int(offset)

    def _query_api(self, params: Dict[str, Union[str, int]]) -> RequestsResponse:
        """Handle actual query to GB API.

        Args:
            params: All requests and required resource query parameters

        Returns:
            The raw requests Response from the GB call
        """
        return get(
            self.URI_BASE + self.RESOURCE_NAME, params=params, headers=self._headers
        )



"""Client for the Games resource of GiantBomb.

https://www.giantbomb.com/api/documentation#toc-0-17
"""
from typing import Any, Dict, Optional, Union

from pybomb.clients.base.client import ResponseParam
from pybomb.clients.base.search_client import SearchClient
from pybomb.response import Response


class GamesClient(SearchClient):
    """Client for the 'games' API resource."""

    RESOURCE_NAME = "games"

    RESPONSE_FIELD_MAP = {
        "aliases": ResponseParam(True, False),
        "api_detail_url": ResponseParam(False, False),
        "date_added": ResponseParam(True, True),
        "date_last_updated": ResponseParam(True, True),
        "deck": ResponseParam(False, False),
        "description": ResponseParam(False, False),
        "expected_release_month": ResponseParam(True, False),
        "expected_release_quarter": ResponseParam(True, False),
        "expected_release_year": ResponseParam(True, False),
        "id": ResponseParam(True, True),
        "image": ResponseParam(False, False),
        "name": ResponseParam(True, True),
        "number_of_user_reviews": ResponseParam(True, True),
        "original_game_rating": ResponseParam(False, True),
        "original_release_date": ResponseParam(True, True),
        "platforms": ResponseParam(True, False),
        "site_detail_url": ResponseParam(False, False),
    }

    def quick_search(
        self,
        name: str,
        sort_by: Optional[str] = None,
        desc: bool = True,
        platform: Optional[int] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        """Search with a simplier API.

        This method allows you to search for a game using only the title and
        the platform.

        Args:
            name: The name of the game to search for
            platform: The platform ID that the game is required to have.
                When set to None no platform filters will be added. Defaults to None
            sort_by: The field to sort the items in the reponse by.
                These will be validated against the availiable sort fields.
            desc: If sort direction is DESC or not (ASC). Defaults to True
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
             A PyBomb Response containg the results of the search
        """
        filter_by: Dict[str, Union[str, int]] = {"name": name}
        if platform is not None:
            filter_by["platforms"] = platform

        return self.search(filter_by=filter_by, sort_by=sort_by, desc=desc)



"""Client for the Platforms resource of GiantBomb.

https://www.giantbomb.com/api/documentation#toc-0-30
"""
from typing import Any, Optional

from pybomb.clients.base.client import ResponseParam
from pybomb.clients.base.search_client import SearchClient
from pybomb.response import Response


class PlatformsClient(SearchClient):
    """Client for the 'platforms' API resource."""

    RESOURCE_NAME = "platforms"

    RESPONSE_FIELD_MAP = {
        "abbreviation": ResponseParam(True, True),
        "api_detail_url": ResponseParam(False, False),
        "company": ResponseParam(True, False),
        "date_added": ResponseParam(True, True),
        "date_last_updated": ResponseParam(True, True),
        "deck": ResponseParam(False, False),
        "description": ResponseParam(False, False),
        "guid": ResponseParam(False, False),
        "id": ResponseParam(True, True),
        "image": ResponseParam(False, False),
        "image_tags": ResponseParam(False, False),
        "install_base": ResponseParam(True, True),
        "name": ResponseParam(True, True),
        "online_support": ResponseParam(True, True),
        "original_price": ResponseParam(True, True),
        "release_date": ResponseParam(True, True),
        "site_detail_url": ResponseParam(False, False),
    }

    def quick_search(
        self,
        name: str,
        sort_by: Optional[str] = None,
        desc: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        """Search with a simplier API.

        This method allows you to search for a platform using only the name.

        Args:
            name: The name of the platform to search for
            sort_by: The field to sort the items in the reponse by.
                These will be validated against the availiable sort fields.
            desc: If sort direction is DESC or not (ASC). Defaults to True
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
             A PyBomb Response containg the results of the search
        """
        return self.search(filter_by={"name": name}, sort_by=sort_by, desc=desc)



"""Client for the Game resource of GiantBomb.

https://www.giantbomb.com/api/documentation#toc-0-16
"""
from typing import Dict, List, Union

from pybomb.clients.base.client import ResponseParam
from pybomb.clients.base.fetch_client import FetchClient
from pybomb.response import Response


class GameClient(FetchClient):
    """Client for the 'game' API resource."""

    RESOURCE_NAME = "game"

    RESPONSE_FIELD_MAP = {
        "aliases": ResponseParam(True, False),
        "api_detail_url": ResponseParam(True, False),
        "characters": ResponseParam(True, False),
        "concepts": ResponseParam(True, False),
        "date_added": ResponseParam(True, False),
        "date_last_updated": ResponseParam(True, False),
        "deck": ResponseParam(True, False),
        "description": ResponseParam(True, False),
        "developers": ResponseParam(True, False),
        "expected_release_day": ResponseParam(True, False),
        "expected_release_month": ResponseParam(True, False),
        "expected_release_quarter": ResponseParam(True, False),
        "expected_release_year": ResponseParam(True, False),
        "first_appearance_characters": ResponseParam(True, False),
        "first_appearance_concepts": ResponseParam(True, False),
        "first_appearance_locations": ResponseParam(True, False),
        "first_appearance_objects": ResponseParam(True, False),
        "first_appearance_people": ResponseParam(True, False),
        "franchises": ResponseParam(True, False),
        "genres": ResponseParam(True, False),
        "id": ResponseParam(True, False),
        "image": ResponseParam(True, False),
        "images": ResponseParam(True, False),
        "killed_characters": ResponseParam(True, False),
        "locations": ResponseParam(True, False),
        "name": ResponseParam(True, False),
        "number_of_user_reviews": ResponseParam(True, False),
        "objects": ResponseParam(True, False),
        "original_game_rating": ResponseParam(False, True),
        "original_release_date": ResponseParam(True, False),
        "people": ResponseParam(True, False),
        "platforms": ResponseParam(True, False),
        "publishers": ResponseParam(True, False),
        "releases": ResponseParam(True, False),
        "reviews": ResponseParam(True, False),
        "similar_games": ResponseParam(True, False),
        "site_detail_url": ResponseParam(True, False),
        "themes": ResponseParam(True, False),
        "videos": ResponseParam(True, False),
    }

    def fetch(self, id_: int, return_fields: List = None) -> Response:
        """Fetch details of a game by ID.

        Args:
            id_: The ID of the game
            return_fields: A list of fields to be returned by the response.
                These will be validated against the availiable return fields.
                The default is to return everything

        Returns:
            A Pybomb Respose object containing the details of the requested game
        """
        game_params: Dict[str, Union[str, int]] = {"id": id_}

        if return_fields is not None:
            self._validate_return_fields(return_fields)
            field_list = ",".join(return_fields)

            game_params["field_list"] = field_list

        response = self._query(game_params)

        return response



"""Exceptions used in this library."""


class PybombException(Exception):
    """Base Exception for library."""

    pass


class ClientException(PybombException):
    """Base Client Exception for module."""

    pass


class InvalidReturnFieldException(ClientException):
    """Exception for invalid return fields."""

    pass


class InvalidSortFieldException(ClientException):
    """Exception for invalid sort fields."""

    pass


class InvalidFilterFieldException(ClientException):
    """Exception for invalid filter fields."""

    pass


class InvalidResponseException(ClientException):
    """Exception thrown when receiving an invalid response from selected resource."""

    pass


class BadRequestException(ClientException):
    """Exception thrown when attempting to send a bad request."""

    pass



"""A library of clients for the main resources of the GiantBomb API.

http://www.giantbomb.com/api/documentation#toc-0-1
"""
try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore

from pybomb.clients.game_client import GameClient
from pybomb.clients.games_client import GamesClient
from pybomb.clients.platforms_client import PlatformsClient
from pybomb.exceptions import (
    BadRequestException,
    ClientException,
    InvalidFilterFieldException,
    InvalidResponseException,
    InvalidReturnFieldException,
    InvalidSortFieldException,
    PybombException,
)

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

# Platform ID's
PS1 = 22
PS2 = 19
PS3 = 35
PS4 = 146

XBOX = 32
XBOX_360 = 20
XBOX_ONE = 145

MAC = 17
PC = 94



"""The response types and factories for PyBomb."""
from typing import NamedTuple, Optional

from requests import Response as RequestsResponse


class Response(NamedTuple):
    """An API response."""

    uri: str
    num_page_results: int
    num_total_results: int
    results: list
    result: Optional[dict]

    @classmethod
    def from_response_data(cls, response_data: RequestsResponse) -> "Response":
        """Response factory.

        If the response is from a fetch client and "results" is not wrapped in
        a list, the single result will be stored in results in a list and the
        original result will be added to result.

        If the response is from a search client and "results" is a list, this list
        will be stored in results and result will be empty.

        Either way, results will always hold the results in list format.

        Args:
            response_data: Raw request response of API call

        Returns:
            A PyBomb Response containing the raw response data.
        """
        response_json = response_data.json()

        cls_args = [
            response_data.url,
            response_json["number_of_page_results"],
            response_json["number_of_total_results"],
        ]

        if isinstance(response_json["results"], dict):
            cls_args.extend(([response_json["results"]], response_json["results"]))
        else:
            cls_args.extend((response_json["results"], None))

        return cls(*cls_args)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pybomb', 'pybomb.clients', 'pybomb.clients.base']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.19.0,<3.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.7.0,<2.0.0']}

setup_kwargs = {
    'name': 'pybomb',
    'version': '0.3.0',
    'description': 'Simple clients for the Giant Bomb API.',
    'long_description': "[![ko-fi](https://www.ko-fi.com/img/githubbutton_sm.svg)](https://ko-fi.com/C0C826VYD)\n\n![Tests](https://github.com/steveYeah/PyBomb/workflows/Tests/badge.svg)\n![Coverage](https://github.com/steveYeah/PyBomb/workflows/Coverage/badge.svg)\n![Release Drafter](https://github.com/steveYeah/PyBomb/workflows/Release%20Drafter/badge.svg)\n![TestPyPi](https://github.com/steveYeah/PyBomb/workflows/TestPyPi/badge.svg)\n![Release](https://github.com/steveYeah/PyBomb/workflows/Release/badge.svg)\n\n[![Codecov](https://codecov.io/gh/steveYeah/PyBomb/branch/master/graph/badge.svg)](https://codecov.io/gh/steveYeah/PyBomb)\n[![PyPI](https://img.shields.io/pypi/v/PyBomb.svg)](https://pypi.org/project/PyBomb/)\n[![Read the Docs](https://readthedocs.org/projects/pybomb/badge/)](https://pybomb.readthedocs.io/)\n\n# PyBomb\n\n>\n\nThis will go into version 1.0 when all resources are supported.\n\n**Currently Supported Resources**:\n\n  - games\n  - game\n  - platforms\n\n## Support OSS, and me :)\nIf you find this project useful, please feel free to [buy me a coffee](https://ko-fi.com/steveyeah)\n\n## Install\n\n``` shell\npip install pybomb\n```\n\n## Examples\n\nTo see a working example of Pybomb, take a look at the example project\n[PyBomb-demo](https://github.com/steveYeah/PyBomb-demo)\n\n**GamesClient - search**\n\n    import pybomb\n\n    my_key = your_giant_bomb_api_key\n    games_client = pybomb.GamesClient(my_key)\n\n    return_fields = ('id', 'name', 'platforms')\n    limit = 10\n    offset = 5\n    sort_by = 'name'\n    filter_by = {'platforms': pybomb.PS3}\n\n    response = games_client.search(\n       filter_by=filter_by,\n       return_fields=return_fields,\n       sort_by=sort_by,\n       desc=True,\n       limit=limit,\n       offset=offset\n    )\n\n    print response.results\n    print response.uri\n    print response.num_page_results\n    print response.num_total_results\n\n**GamesClient - quick\\_search**\n\n    import pybomb\n\n    my_key = your_giant_bomb_api_key\n    games_client = pybomb.GamesClient(my_key)\n\n    response = games_client.quick_search(\n      name='call of duty',\n      platform=pybomb.PS3,\n      sort_by='original_release_date',\n      desc=True\n    )\n\n    print response.results\n    print response.uri\n    print response.num_page_results\n    print response.num_total_results\n\n## Documentation\n\nThe full documentation, including more examples can be found at\n[readthedocs](https://pybomb.readthedocs.org)\n\n## Contributing\n\nThis project uses [Poetry](https://python-poetry.org/) and [Nox](https://nox.thea.codes/en/stable/) so make sure you have those setup!\n\nWhilst working you can use\n\n```shell\n$ nox -rs tests\n```\nto run the tests, but once you have finished, make sure to run all of nox before making a PR\n\n```shell\n$ nox\n```\n\nThis project also uses [Pytest](https://docs.pytest.org/en/stable/) for tests, and we aim for 100% test coverage. The build will fail when the coverage is less than this, but feel free to use `# pragma: no cover` if it makes sense to do so (although this should be very rare). [Mypy](http://mypy-lang.org/) is also used, and we even use [typeguard](https://typeguard.readthedocs.io/en/latest/) so typing is important here!\n\nI am always happy to help where I can and I try to be as responsive as possible to PRs. Email me if I am slow, or if you need any help!\n",
    'author': 'steveYeah',
    'author_email': 'hutchinsteve@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/steveYeah/PyBomb',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""Base client to extend to create fetch clients for endpoints of the GiantBomb API."""
from abc import abstractmethod
from typing import Dict, List, Union

from requests import get, Response as RequestsResponse

from pybomb.clients.base.client import Client
from pybomb.response import Response


class FetchClient(Client):
    """Base class for fetch GB API resource clients."""

    @abstractmethod
    def fetch(self, id_: int, return_fields: List = None) -> Response:
        """Fetch details of a game by ID."""
        ...  # pragma: no cover

    def _query_api(self, params: Dict[str, Union[str, int]]) -> RequestsResponse:
        """Handle actual query to GB API.

        Args:
            params: All requests and required resource query parameters

        Returns:
            The raw requests Response from the GB call
        """
        id = params.pop("id")
        return get(
            f"{self.URI_BASE}{self.RESOURCE_NAME}/{id}",
            params=params,
            headers=self._headers,
        )
